import logging

from .flumine import Flumine
from .backtest.backtest import FlumineBacktest
from .strategy.strategy import BaseStrategy
from .exceptions import FlumineException
from .__version__ import __title__, __version__, __author__

from betfairlightweight.resources import bettingresources
from .patching import EX, SP

# patch bflw with faster classes
bettingresources.RunnerBookEX = EX
bettingresources.RunnerBookSP = SP

# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger(__name__).addHandler(logging.NullHandler())



import unittest
from unittest import mock

from flumine import config
from flumine.order.orderpackage import (
    OrderPackageType,
    BaseOrderPackage,
    EventType,
    QueueType,
    BetfairOrderPackage,
    ExchangeType,
    OrderStatus,
)


class OrderPackageTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_package_type = mock.Mock()
        self.mock_client = mock.Mock()
        self.mock_order = mock.Mock()
        self.mock_order.status = OrderStatus.PENDING
        self.order_package = BaseOrderPackage(
            self.mock_client,
            "1.234",
            [self.mock_order],
            self.mock_package_type,
            1,
            market_version=123,
        )

    def test_init(self):
        self.assertEqual(self.order_package.client, self.mock_client)
        self.assertEqual(self.order_package.market_id, "1.234")
        self.assertEqual(self.order_package._orders, [self.mock_order])
        self.assertEqual(self.order_package.package_type, self.mock_package_type)
        self.assertEqual(self.order_package.EVENT_TYPE, EventType.ORDER_PACKAGE)
        self.assertEqual(self.order_package.QUEUE_TYPE, QueueType.HANDLER)
        self.assertEqual(self.order_package.bet_delay, 1)
        self.assertIsNone(self.order_package.EXCHANGE)
        self.assertFalse(self.order_package.async_)
        self.assertEqual(self.order_package._market_version, 123)
        self.assertFalse(self.order_package.processed)
        self.assertTrue(self.order_package._retry)
        self.assertEqual(self.order_package._max_retries, 3)
        self.assertEqual(self.order_package._retry_count, 0)
        self.assertIsNone(self.order_package.simulated_delay)

    def test_retry(self):
        self.assertTrue(self.order_package.retry())
        self.assertEqual(self.order_package._retry_count, 1)

    @mock.patch("flumine.order.orderpackage.time")
    def test_retry_false_count(self, mock_time):
        self.assertTrue(self.order_package.retry())
        self.assertEqual(self.order_package._retry_count, 1)
        self.assertTrue(self.order_package.retry())
        self.assertEqual(self.order_package._retry_count, 2)
        self.assertTrue(self.order_package.retry())
        self.assertEqual(self.order_package._retry_count, 3)
        self.assertFalse(self.order_package.retry())
        self.assertEqual(self.order_package._retry_count, 3)
        mock_time.sleep.assert_called()

    def test_reset_orders(self):
        mock_order = mock.MagicMock()
        self.order_package._orders = [mock_order]
        self.order_package.reset_orders()
        mock_order.executable.assert_called()
        mock_order.trade.__enter__.assert_called()
        self.order_package.reset_orders(True)
        mock_order.execution_complete.assert_called()

    def test_calc_simulated_delay(self):
        config.place_latency = 0.1
        config.cancel_latency = 0.2
        config.update_latency = 0.3
        config.replace_latency = 0.4

        self.assertIsNone(self.order_package.calc_simulated_delay())
        self.order_package.client.execution.EXCHANGE = ExchangeType.SIMULATED
        self.order_package.package_type = OrderPackageType.PLACE
        self.assertEqual(self.order_package.calc_simulated_delay(), 1.1)
        self.order_package.package_type = OrderPackageType.CANCEL
        self.assertEqual(self.order_package.calc_simulated_delay(), 0.2)
        self.order_package.package_type = OrderPackageType.UPDATE
        self.assertEqual(self.order_package.calc_simulated_delay(), 0.3)
        self.order_package.package_type = OrderPackageType.REPLACE
        self.assertEqual(self.order_package.calc_simulated_delay(), 1.4)

    def test_place_instructions(self):
        with self.assertRaises(NotImplementedError):
            assert self.order_package.place_instructions

    def test_cancel_instructions(self):
        with self.assertRaises(NotImplementedError):
            assert self.order_package.cancel_instructions

    def test_update_instructions(self):
        with self.assertRaises(NotImplementedError):
            assert self.order_package.update_instructions

    def test_replace_instructions(self):
        with self.assertRaises(NotImplementedError):
            assert self.order_package.replace_instructions

    def test_order_limit(self):
        with self.assertRaises(NotImplementedError):
            assert self.order_package.order_limit(OrderPackageType.PLACE)

    def test_orders(self):
        self.assertEqual(self.order_package.orders, [self.mock_order])
        self.order_package._orders = [
            mock.Mock(status=OrderStatus.PENDING),
            mock.Mock(status=OrderStatus.PENDING),
            mock.Mock(status=OrderStatus.VIOLATION),
        ]
        self.assertEqual(len(self.order_package.orders), 2)

    def test_retry_count(self):
        self.order_package._retry_count = 1
        self.assertEqual(self.order_package.retry_count, 1)

    def test_info(self):
        self.assertEqual(
            self.order_package.info,
            {
                "id": self.order_package.id,
                "client": self.order_package.client,
                "market_id": self.order_package.market_id,
                "orders": [self.mock_order.id],
                "order_count": 1,
                "package_type": self.order_package.package_type.value,
                "customer_strategy_ref": self.order_package.customer_strategy_ref,
                "bet_delay": self.order_package.bet_delay,
                "market_version": self.order_package._market_version,
                "retry": self.order_package._retry,
                "retry_count": self.order_package._retry_count,
                "async": self.order_package.async_,
            },
        )

    def test_market_version(self):
        self.assertEqual(
            self.order_package.market_version,
            {"version": self.order_package._market_version},
        )

    def test_iter(self):
        self.assertEqual([i for i in self.order_package], self.order_package.orders)

    def test_len(self):
        self.assertEqual(len(self.order_package), 1)


class BetfairOrderPackageTest(unittest.TestCase):
    def setUp(self) -> None:
        self.mock_package_type = mock.Mock()
        self.mock_client = mock.Mock()
        self.mock_order = mock.Mock()
        self.mock_order.status = OrderStatus.PENDING
        self.order_package = BetfairOrderPackage(
            self.mock_client,
            "1.234",
            [self.mock_order],
            self.mock_package_type,
            0,
        )

    def test_init(self):
        self.assertEqual(self.order_package.EXCHANGE, ExchangeType.BETFAIR)

    def test_place_instructions(self):
        self.assertEqual(
            self.order_package.place_instructions,
            [self.mock_order.create_place_instruction()],
        )

    def test_cancel_instructions(self):
        self.assertEqual(
            self.order_package.cancel_instructions,
            [self.mock_order.create_cancel_instruction()],
        )

    def test_update_instructions(self):
        self.assertEqual(
            self.order_package.update_instructions,
            [self.mock_order.create_update_instruction()],
        )

    def test_replace_instructions(self):
        self.assertEqual(
            self.order_package.replace_instructions,
            [self.mock_order.create_replace_instruction()],
        )

    def test_order_limit(self):
        self.assertEqual(self.order_package.order_limit(OrderPackageType.PLACE), 200)
        self.assertEqual(self.order_package.order_limit(OrderPackageType.CANCEL), 60)
        self.assertEqual(self.order_package.order_limit(OrderPackageType.UPDATE), 60)
        self.assertEqual(self.order_package.order_limit(OrderPackageType.REPLACE), 60)



import logging
from collections import defaultdict

from ..order.orderpackage import OrderPackageType, BetfairOrderPackage
from ..events import events
from ..exceptions import ControlError, OrderError
from ..utils import chunks, get_market_notes

logger = logging.getLogger(__name__)


class Transaction:
    """
    Process place, cancel, update and replace requests.

    Default behaviour is to execute immediately however
    when it is used as a context manager requests can
    be batched, for example:

        with market.transaction() as t:
            market.place_order(order)  # executed immediately in separate transaction
            t.place_order(order)  # executed on transaction __exit__

        with market.transaction() as t:
            t.place_order(order)
            ..
            t.execute()  # above order executed
            ..
            t.cancel_order(order)
            t.place_order(order)  # both executed on transaction __exit__
    """

    def __init__(self, market, id_: int, async_place_orders: bool):
        self.market = market
        self._id = id_  # unique per market only
        self._async_place_orders = async_place_orders
        self._pending_orders = False
        self._pending_place = []  # list of (<Order>, market_version)
        self._pending_cancel = []  # list of (<Order>, None)
        self._pending_update = []  # list of (<Order>, None)
        self._pending_replace = []  # list of (<Order>, market_version)

    def place_order(
        self,
        order,
        market_version: int = None,
        execute: bool = True,
        force: bool = False,
    ) -> bool:
        if (
            execute
            and not force
            and self._validate_controls(order, OrderPackageType.PLACE) is False
        ):
            return False
        # place
        order.place(
            self.market.market_book.publish_time,
            market_version,
            self._async_place_orders,
        )
        if self.market.blotter.has_trade(order.trade.id):
            new_trade = False
        else:
            new_trade = True
        if order.id not in self.market.blotter:
            self.market.blotter[order.id] = order
            # update market_notes
            market_notes = get_market_notes(self.market, order.selection_id)
            order.market_notes = market_notes
            if order.trade.market_notes is None:
                order.trade.market_notes = market_notes
        else:
            raise OrderError("Order %s has already been placed" % order.id)
        if new_trade:
            self.market.flumine.log_control(events.TradeEvent(order.trade))
        if execute:  # handles replaceOrder
            runner_context = order.trade.strategy.get_runner_context(*order.lookup)
            runner_context.place(order.trade.id)
            self._pending_place.append((order, market_version))
            self._pending_orders = True
        return True

    def cancel_order(
        self, order, size_reduction: float = None, force: bool = False
    ) -> bool:
        if (
            not force
            and self._validate_controls(order, OrderPackageType.CANCEL) is False
        ):
            return False
        # cancel
        order.cancel(size_reduction)
        self._pending_cancel.append((order, None))
        self._pending_orders = True
        return True

    def update_order(
        self, order, new_persistence_type: str, force: bool = False
    ) -> bool:
        if (
            not force
            and self._validate_controls(order, OrderPackageType.UPDATE) is False
        ):
            return False
        # update
        order.update(new_persistence_type)
        self._pending_update.append((order, None))
        self._pending_orders = True
        return True

    def replace_order(
        self, order, new_price: float, market_version: int = None, force: bool = False
    ) -> bool:
        if (
            not force
            and self._validate_controls(order, OrderPackageType.REPLACE) is False
        ):
            return False
        # replace
        order.replace(new_price)
        self._pending_replace.append((order, market_version))
        self._pending_orders = True
        return True

    def execute(self) -> int:
        packages = []
        if self._pending_place:
            packages += self._create_order_package(
                self._pending_place,
                OrderPackageType.PLACE,
                async_=self._async_place_orders,
            )
        if self._pending_cancel:
            packages += self._create_order_package(
                self._pending_cancel,
                OrderPackageType.CANCEL,
            )
        if self._pending_update:
            packages += self._create_order_package(
                self._pending_update,
                OrderPackageType.UPDATE,
            )
        if self._pending_replace:
            packages += self._create_order_package(
                self._pending_replace,
                OrderPackageType.REPLACE,
            )
        if packages:
            for package in packages:
                self.market.flumine.process_order_package(package)
            if logger.isEnabledFor(logging.INFO):
                logger.info(
                    "%s order packages executed in transaction" % len(packages),
                    extra={
                        "market_id": self.market.market_id,
                        "order_packages": [o.info for o in packages],
                        "transaction_id": self._id,
                    },
                )
            self._pending_orders = False
        return len(packages)

    def _validate_controls(self, order, package_type: OrderPackageType) -> bool:
        # return False on violation
        try:
            for control in self.market.flumine.trading_controls:
                control(order, package_type)
            for control in self.market.flumine.client.trading_controls:
                control(order, package_type)
        except ControlError:
            return False
        else:
            return True

    def _create_order_package(
        self, orders: list, package_type: OrderPackageType, async_: bool = False
    ) -> list:
        # group orders by marketVersion
        orders_grouped = defaultdict(list)
        for o in orders:
            orders_grouped[o[1]].append(o[0])
        # create packages (chunked by limit)
        limit = BetfairOrderPackage.order_limit(package_type)
        packages = []
        for market_version, package_orders in orders_grouped.items():
            for chunked_orders in chunks(package_orders, limit):
                packages.append(
                    BetfairOrderPackage(
                        client=self.market.flumine.client,
                        market_id=self.market.market_id,
                        orders=chunked_orders,
                        package_type=package_type,
                        bet_delay=self.market.market_book.bet_delay,
                        market_version=market_version,
                        async_=async_,
                    )
                )
        orders.clear()
        return packages

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._pending_orders:
            self.execute()
